import traceback

from copy import deepcopy as copy

import bpy
import bmesh

from math import degrees, radians, floor
from mathutils import Vector, Matrix
from bgl import *

from .. draw_px import anchor_line, text, dot

from ... utils.math import vector_sum
from ... utils.blender_ui import dpi, dpi_factor
from ... utils.space_3d import mouse_intersect3d, location3d_to_point2d, point2d_to_location3d

from ... preferences import preference

class Shape:
    traceback = lambda: print('\n', traceback.format_exc())

    #TODO: use this to store last state globally when 2d is deprecated
    #mode = 'CUT' # SLICE, MAKE, JOIN
    verts = list()
    edges = list()
    faces = list()

    hide = bool()
    display_bbox = True
    display_negative = True

    points = {
        'front':  (1, 3, 5, 7),
        'back':   (0, 2, 4, 6),
        'left':   (4, 5, 6, 7),
        'right':  (0, 1, 2, 3),
        'top':    (2, 3, 6, 7),
        'bottom': (0, 1, 4, 5)}

    bbox_edges = [(1, 0), (0, 3), (3, 2), (2, 1),
                  (7, 3), (2, 6), (6, 7), (4, 7),
                  (6, 5), (5, 4), (0, 4), (5, 1)]

    bbox_faces = [(0, 1, 2, 3), (3, 2, 6, 7), (7, 6, 5, 4),
                  (4, 5, 1, 0), (3, 7, 4, 0), (6, 2, 1, 5)]

    def __init__(self, op, context):
        self.create(op, context)

    @staticmethod
    def matrix(op, coord, local=False):
        return op.lat.matrix_world * coord if not local else op.lat.matrix_world.inverted() * coord

    @staticmethod
    def center(op, local=False):
        points = (Vector(op.lat.data.points[point].co_deform[:]) for point in (0, 1, 2, 3, 4, 5, 6, 7))
        return Shape.matrix(op, 0.125 * vector_sum(points), local)

    @staticmethod
    def side_center(op, points, local=False):
        points = (Vector(op.lat.data.points[point].co_deform[:]) for point in points)
        return Shape.matrix(op, 0.25 * vector_sum(points), local)

    @staticmethod
    def edge_center(op, points, local=False):
        points = (Vector(op.lat.data.points[point].co_deform[:]) for point in points)
        return Shape.matrix(op, 0.5 * vector_sum(points), local)

    @staticmethod
    def point(op, point, local=False):
        return Shape.matrix(op, Vector(op.lat.data.points[point].co_deform[:]), local)

    class create:

        def __init__(self, op, context):
            try:
                self.shape(op, context)
                self.lattice(op, context)
            except Exception: Shape.traceback()

        @staticmethod
        def shape(op, context):
            mesh = bpy.data.meshes.new('BC-shape')
            mesh.from_pydata(op.verts, op.edges, op.faces)

            op.shape = bpy.data.objects.new('BC-shape', mesh)
            op.shape.select = True

            track_quat = op.ray_object.data.polygons[op.face_index].normal.to_track_quat('Z', 'Y')
            track_mat = track_quat.to_matrix().to_4x4()
            track_mat.translation = op.ray_object.data.polygons[op.face_index].center

            op.shape.matrix_world = op.ray_object.matrix_world * track_mat
            op.shape.matrix_world.translation = op.hit_location
            op.origin = op.shape.matrix_world.translation

            #TODO: replace with generated normal
            op.depth_normal = op.shape.matrix_world.to_3x3() * mesh.polygons[1].normal

            op.shape.hide = not Shape.hide
            op.shape.show_wire = True
            op.shape.show_all_edges = True

            context.scene.objects.link(op.shape)

        @staticmethod
        def lattice(op, context):
            lat = bpy.data.lattices.new('BC-lattice')
            op.lat = bpy.data.objects.new('BC-lattice', lat)
            op.lat.select = True

            op.lat.matrix_world = op.shape.matrix_world
            op.lat.matrix_world.translation = op.shape.matrix_world.translation
            op.lat.hide = not Shape.hide
            op.lat.data.interpolation_type_u = 'KEY_LINEAR'
            op.lat.data.interpolation_type_v = 'KEY_LINEAR'
            op.lat.data.interpolation_type_w = 'KEY_LINEAR'

            context.scene.objects.link(op.lat)
            context.scene.update()

            op.latmod = op.shape.modifiers.new('BC-lat', 'LATTICE')
            op.latmod.object = op.lat

            vector = Vector((0.000001, 0.000001, 0.000001))
            for point in op.lat.data.points:
                point.co_deform = vector

            for point in Shape.points['front']:
                op.lat.data.points[point].co_deform.z += 0.001

    def draw2d(op, context):
        try:
            if op.modified:
                Shape.update.widgets(op, context)

                for mode in op.widget_location:
                    dot(op.widget_location[mode], color=op.widget_color[mode])

                if op.shape_mode == "ROTATE":
                    anchor_line(op.pivot_point, op.mouse)
                    text(op.pivot_point, '{}: {}'.format(op.axis_mode, op.angle_text))

                elif op.shape_mode == "BEVEL":
                    if op.bevmod:
                        text(Vector((op.mouse.x + 30, op.mouse.y)), str(op.bevmod.segments))
                        text(location3d_to_point2d(context, op.origin), '{:.4f}'.format(op.bevmod.width))

                elif op.shape_mode == "ARRAY":
                    text(Vector((op.mouse.x + 30, op.mouse.y)), str(op.arrmod.count))
                    text(location3d_to_point2d(context, op.origin), '{}: {:.4f}'.format(op.axis_mode, getattr(op.arrmod, 'relative_offset_displace')[op.axis_index]))

                elif op.shape_mode == "SOLIDIFY":
                    text(Vector((op.mouse.x + 30, op.mouse.y)), '{:.4f}'.format(op.solmod.thickness))

        except Exception: Shape.traceback()

    class draw3d:
        __slots__ = ()

        def __init__(self, op, context):
            try:
                Shape.update.geometry(op, context)

                if not Shape.hide and not op.draw_cutoff:
                    self.geometry(op, Shape.face_color(op.mode), Shape.edge_color(op.shape_mode))
                    if Shape.display_negative:
                        self.geometry(op, Shape.face_color.negative, None, negative=True)
                    if Shape.display_bbox:
                        self.bbox(op, Shape.face_color.bbox, Shape.edge_color.bbox)

            except Exception: Shape.traceback()

        @staticmethod
        def cutoff(op):
            if op.arrmod:
                draw_cutoff = 200

                if op.bevmod:
                    draw_cutoff = 40 if op.bevel_mode != 'ROUND' else 20 # TODO: additionally adjust based on segments

                if op.mirmod:
                    if op.mirmod.use_x:
                        draw_cutoff *= 0.5
                    if op.mirmod.use_y:
                        draw_cutoff *= 0.5
                    if op.mirmod.use_z:
                        draw_cutoff *= 0.5

                if not Shape.display_negative:
                    draw_cutoff *= 2

                if op.arrmod.count > draw_cutoff:
                    if not op.draw_cutoff:
                        op.report({'WARNING'}, 'Too much geometry to draw, displaying shape only')

                    op.draw_cutoff = True
                    op.lat.hide = False
                    op.shape.hide = False

                elif op.draw_cutoff:
                    op.draw_cutoff = False

                    op.lat.hide = not op.lat.hide
                    op.shape.hide = not op.shape.hide

        @staticmethod
        def geometry(op, face_color, edge_color, negative=False):
            glEnable(GL_BLEND)

            if not negative: glEnable(GL_DEPTH_TEST)

            glEnable(GL_CULL_FACE)

            glColor4f(*face_color)

            #TODO: add toggle (F-key) to flip draw direction
            glCullFace(GL_BACK)

            for face in Shape.faces:
                glBegin(GL_POLYGON)

                for vert in face:
                    glVertex3f(*Shape.matrix(op, vert))

                glEnd()

            glDisable(GL_CULL_FACE)

            if not negative:
                glColor4f(*edge_color)

                # TODO: add user preference for x-ray wire
                if op.shape_mode in {'BEVEL', 'SOLIDIFY', 'ARRAY'}:
                    glDisable(GL_DEPTH_TEST)

                for edge in Shape.edges:
                    glBegin(GL_LINES)

                    for vert in edge:
                        glVertex3f(*Shape.matrix(op, vert))

                    glEnd()

            if not negative: glDisable(GL_DEPTH_TEST)

            glDisable(GL_BLEND)

        @staticmethod
        def bbox(op, face_color, edge_color):
            glEnable(GL_BLEND)

            bound_box = op.shape.bound_box

            glColor4f(*face_color)

            for index in range(6):
                glBegin(GL_POLYGON)

                glVertex3f(*Shape.matrix(op, Vector(bound_box[Shape.bbox_faces[index][0]][:])))
                glVertex3f(*Shape.matrix(op, Vector(bound_box[Shape.bbox_faces[index][1]][:])))
                glVertex3f(*Shape.matrix(op, Vector(bound_box[Shape.bbox_faces[index][2]][:])))
                glVertex3f(*Shape.matrix(op, Vector(bound_box[Shape.bbox_faces[index][3]][:])))

                glEnd()

            glEnable(GL_LINE_STIPPLE)

            glColor4f(*edge_color)

            for index in range(12):
                glBegin(GL_LINES)

                glVertex3f(*Shape.matrix(op, Vector(bound_box[Shape.bbox_edges[index][0]][:])))
                glVertex3f(*Shape.matrix(op, Vector(bound_box[Shape.bbox_edges[index][1]][:])))

                glEnd()

            glDisable(GL_LINE_STIPPLE)
            glDisable(GL_BLEND)

    class update:
        __slots__ = ()

        def __init__(self, op, context, event):
            try:
                if op.shape_mode:
                    getattr(self, op.shape_mode)(op, context, event)
                if not op.mirmod:
                    Shape.update.mirror(op, context, event)

            except Exception: Shape.traceback()

        @staticmethod
        def geometry(op, context):
            Shape.draw3d.cutoff(op)
            Shape.update.bevel_clamp_overlap(op)

            if not Shape.hide and op.update:
                mesh = op.shape.to_mesh(context.scene, apply_modifiers=True, settings='PREVIEW', calc_tessface=False)
                mesh.validate()

                bm = bmesh.new(use_operators=False)
                bm.from_mesh(mesh)

                # Shape.verts = tuple([vert.co for vert in bm.verts])
                Shape.edges = tuple([tuple([copy(vert.co) for vert in edge.verts]) for edge in bm.edges])
                Shape.faces = tuple([tuple([copy(vert.co) for vert in face.verts]) for face in bm.faces])

                try: bpy.data.meshes.remove(mesh)
                except Exception: Shape.traceback()

                op.update = False

        @staticmethod
        def widgets(op, context):
            modes = ('DRAW', 'EXPAND', 'OFFSET')
            locations = (
                location3d_to_point2d(context, Shape.matrix(op, op.lat.data.points[op.draw_dot_point].co_deform)), # draw
                location3d_to_point2d(context, Shape.side_center(op, Shape.points['back'])), # expand
                location3d_to_point2d(context, Shape.side_center(op, Shape.points['front']))) # offset

            for index, mode in enumerate(modes):
                op.widget_location[mode] = locations[index]
                op.widget_color[mode] = Shape.widget_color.faded

            closest, mode = min((op.mouse - location, modes[index]) for index, location in enumerate(locations))

            margin = 10
            if -margin < closest.x < margin and -margin < closest.y < margin:
                op.widget_color[mode] = Shape.widget_color(mode)
                op.active_widget = mode
            else: op.active_widget = None

        @staticmethod
        def mirror(op, context, event):
            if not op.mirmod:
                op.mirmod = op.shape.modifiers.new('BC-mirror', 'MIRROR')
                op.mirmod.use_x = False
                op.mirmod.use_mirror_merge = True
                op.mirmod.use_clip = True

                #TODO: dont add empty if not BoxCutter_mirror_type == 'CURSOR'
                #TODO: check to see if there is already an empty at cursor location before adding
                op.empty = bpy.data.objects.new('BC-mirror-object', None)
                op.empty.hide = True
                context.scene.objects.link(op.empty)

                op.mirmod.mirror_object = op.empty

                if preference().BoxCutter_mirror_type == '3D':
                    op.empty.matrix_world = op.ray_object.matrix_world
                else:
                    op.empty.matrix_world = op.ray_object.matrix_world
                    op.empty.location = context.space_data.cursor_location

                for axis in 'xyz':
                    setattr(op.mirmod, 'use_{}'.format(axis), getattr(preference(), 'BoxCutter_mirror_axis_{}'.format(axis)))

                Shape.update.modifier_order(op, context)
            else:
                for axis in 'xyz':
                    setattr(op.mirmod, 'use_{}'.format(axis), getattr(preference(), 'BoxCutter_mirror_axis_{}'.format(axis)))

        @staticmethod
        def modifier_order(op, context):
            modifiers = op.shape.modifiers[:]

            if op.arrmod and op.mirmod:
                index = modifiers.index(op.arrmod)

                context.scene.objects.active = op.shape

                while index > modifiers.index(op.mirmod):
                    bpy.ops.object.modifier_move_up(modifier=op.arrmod.name)
                    index -= 1

                context.scene.objects.active = op.active_object

        @staticmethod
        def bevel_clamp_overlap(op):
            #TODO: replace with generalized method
            if op.bevmod:

                if op.bevmod.segments > 20:
                    op.bevmod.segments = 20

                preference().BoxCutter_bevel_segments = op.bevmod.segments

                if op.bevmod_use_clamp_overlap: # this will not work for all shapes accurately
                    vector1 = Vector(op.lat.bound_box[0][:])
                    vector2 = Vector(op.lat.bound_box[1][:])
                    vector3 = Vector(op.lat.bound_box[5][:])
                    vector4 = Vector(op.lat.bound_box[6][:])
                    distances = [vector4 - vector3, vector3 - vector2]

                    if op.bevel_mode != 'ROLL':
                        distances.append((vector2 - vector1) * 2)

                    min_distance = max(min(distances))

                    if op.bevmod.width > min_distance * 0.5:
                        op.bevmod.width = min_distance * 0.5 - 0.00001

        @staticmethod
        def DRAW(op, context, event):
            mouse_intersect = mouse_intersect3d(context, event, Shape.side_center(op, Shape.points['front']), op.face_normal)
            mouse_intersect = Shape.matrix(op, mouse_intersect, local=True)

            if preference().BoxCutter_enable_grid:
                mouse_intersect.x = float('{:.1f}'.format(mouse_intersect.x))
                mouse_intersect.y = float('{:.1f}'.format(mouse_intersect.y))

            origin_point = Shape.matrix(op, op.origin, local=True)

            side = 0 if mouse_intersect.x > origin_point.x else 1

            sides = ('left', 'right')
            target_side = sides[side]
            clear_side = sides[not side]

            op.draw_dot_point = ((5, 7), (1, 3))[side]
            op.expand_point = ((4, 6), (0, 2))[side]

            if not op.constrain or op.constrain and op.axis_mode == 'X':
                for point in Shape.points[target_side]:
                    op.lat.data.points[point].co_deform.x = mouse_intersect.x

                for point in Shape.points[clear_side]:
                    op.lat.data.points[point].co_deform.x = 0

            side = 0 if mouse_intersect.y > origin_point.y else 1

            sides = ('bottom', 'top')
            target_side = sides[side]
            clear_side = sides[not side]

            op.draw_dot_point = op.draw_dot_point[side]
            op.expand_point = op.expand_point[side]

            if not op.constrain or op.constrain and op.axis_mode == 'Y':
                for point in Shape.points[target_side]:
                    op.lat.data.points[point].co_deform.y = mouse_intersect.y

                for point in Shape.points[clear_side]:
                    op.lat.data.points[point].co_deform.y = 0

            op.update = True

        @staticmethod
        def EXPAND(op, context, event):
            if op.modified:
                location = Shape.side_center(op, Shape.points['back'])
            else:
                location = Shape.matrix(op, op.lat.data.points[op.expand_point].co_deform)

            location = Shape.matrix(op, point2d_to_location3d(context, op.mouse, location), local=True)

            if preference().BoxCutter_enable_grid:
                location.z = float('{:.1f}'.format(location.z))

            for point in Shape.points['back']:
                op.lat.data.points[point].co_deform.z = location.z

            op.update = True

        @staticmethod
        def OFFSET(op, context, event):
            location = Shape.matrix(op, point2d_to_location3d(context, op.mouse, Shape.side_center(op, Shape.points['front'])), local=True)

            if preference().BoxCutter_enable_grid:
                location.z = float('{:.1f}'.format(location.z))

            for point in Shape.points['front']:
                op.lat.data.points[point].co_deform.z = location.z

            op.update = True

        @staticmethod
        def GRAB(op, context, event):
            mouse_intersect = Shape.matrix(op, mouse_intersect3d(context, event, op.origin, op.face_normal), local=True)

            if preference().BoxCutter_enable_grid:
                mouse_intersect.x = float('{:.1f}'.format(mouse_intersect.x))
                mouse_intersect.y = float('{:.1f}'.format(mouse_intersect.y))

            if not op.constrain:
                vector = op.shape.matrix_world.to_3x3() * Vector((mouse_intersect.x, mouse_intersect.y, 0))
                op.lat.matrix_world.translation += vector
                op.shape.matrix_world.translation += vector

            elif op.axis_mode == 'X':
                vector = op.shape.matrix_world.to_3x3() * Vector((mouse_intersect.x, 0, 0))
                op.lat.matrix_world.translation += vector
                op.shape.matrix_world.translation += vector

            elif op.axis_mode == 'Y':
                vector = op.shape.matrix_world.to_3x3() * Vector((0, mouse_intersect.y, 0))
                op.lat.matrix_world.translation += vector
                op.shape.matrix_world.translation += vector

            elif op.axis_mode == 'Z':
                mouse_intersect = mouse_intersect3d(context, event, Shape.side_center(op, Shape.points['front']), op.depth_normal)
                mouse_intersect = Shape.matrix(op, mouse_intersect, local=True)

                if preference().BoxCutter_enable_grid:
                    mouse_intersect.z = float('{:.1f}'.format(mouse_intersect.z))

                vector = op.shape.matrix_world.to_3x3() * Vector((0, 0, mouse_intersect.z))
                op.lat.matrix_world.translation += vector
                op.shape.matrix_world.translation += vector

            op.update = True

        @staticmethod
        def ROTATE(op, context, event):
            mouse = op.mouse - op.pivot_point
            alpha = op.mouse_start.angle_signed(mouse)

            op.angle_track += degrees(alpha)

            angle = preference().BoxCutter_rotation_angle
            if abs(op.angle_track) > angle:
                multiplier = abs(op.angle_track)//angle
                angle = angle * multiplier
                alpha = radians(angle) if op.angle_track > 0 else -radians(angle)
                op.angle += degrees(alpha)
                op.angle_track = 0
            else:
                alpha = 0

            op.angle_text = str(floor(op.angle))

            op.shape.matrix_world *= Matrix.Rotation(-alpha, 4, op.axis_mode)
            op.lat.matrix_world *= Matrix.Rotation(-alpha, 4, op.axis_mode)

            op.mouse_start = mouse

            op.update = True

        @staticmethod
        def BEVEL(op, context, event):
            if not op.bevmod:
                width = (op.mouse.x - op.mouse_start.x) / dpi_factor()
                factor = 0.0001 if event.shift else 0.001
                if (op.bevmod_start_width + width * factor) > 0.00001:

                    op.shape.data.use_customdata_edge_bevel = True

                    for edge in (4, 5, 10, 11):
                        op.shape.data.edges[edge].bevel_weight = 1
                    op.shape.data.validate()

                    op.bevmod = op.shape.modifiers.new('BC-bevel', 'BEVEL')
                    op.bevmod.width = 0.00001
                    op.bevmod.segments = preference().BoxCutter_bevel_segments
                    op.bevmod.use_clamp_overlap = False
                    op.bevmod.limit_method = 'WEIGHT'

            else:
                width = (op.mouse.x - op.mouse_start.x) / dpi_factor()
                factor = 0.0001 if event.shift else 0.001
                op.bevmod.width = op.bevmod_start_width + width * factor

                Shape.update.bevel_clamp_overlap(op)

                if op.bevmod.width < 0.00001:
                    op.shape.modifiers.remove(op.bevmod)
                    op.bevmod = None

            op.update = True

        @staticmethod
        def SOLIDIFY(op, context, event):
            if not op.solmod:
                op.solmod = op.shape.modifiers.new('BC-solidify', 'SOLIDIFY')
                op.solmod.offset = 1
                op.solmod.use_even_offset = True
                op.solmod.use_quality_normals = True

            else:
                thickness = (op.mouse.x - op.mouse_start.x) / dpi_factor()
                factor = 0.0001 if event.shift else 0.001
                op.solmod.thickness = op.solmod_start_thickness + thickness * factor

            op.update = True

        @staticmethod
        def ARRAY(op, context, event):
            if not op.arrmod:
                op.arrmod = op.shape.modifiers.new('BC-array', 'ARRAY')
                op.arrmod.count = op.arrmod_count
                op.arrmod.count = 2 if op.arrmod.count < 2 else op.arrmod.count
                op.arrmod.relative_offset_displace[op.axis_index] = 1

                op.arrmod_start_count = 1

                Shape.update.modifier_order(op, context)

            else:
                for index, offset in enumerate(op.arrmod.relative_offset_displace):
                    if index != op.axis_index:
                        op.arrmod.relative_offset_displace[index] = 0

                op.arrmod.count = 2 if op.arrmod.count < 2 else op.arrmod.count

                offset = (op.mouse.x - op.mouse_start.x) / dpi_factor()
                factor = 0.001 if event.shift else 0.01
                op.arrmod.relative_offset_displace[op.axis_index] = op.arrmod_start_offset - offset * factor

            op.update = True

    #TODO: user preference options for colors
    class face_color:
        negative = (0.1, 0.1, 0.1, 0.2)
        bbox = (0.1, 0.1, 0.1, 0.033)

        def __new__(self, mode):
            return self.corrected(getattr(self, mode)())

        def corrected(color):
            r, b, g, a = color

            factor = 0.0
            if Shape.display_negative and Shape.display_bbox:
                factor += 0.5
            elif Shape.display_negative:
                factor += 0.35
            elif Shape.display_bbox:
                factor += 0.033

            return (r + (r * factor), b + (b * factor), g + (g * factor), a)

        def CUT():
            return preference().BC_redbox_color

        def SLICE():
            return preference().BC_yellowbox_color

        def MAKE():
            return preference().BC_graybox_color

        def JOIN():
            return preference().BC_greenbox_color

    class edge_color:
        default = (0.1, 0.1, 0.1, 0.75)
        bbox = (0.1, 0.1, 0.1, 0.5)

        def __new__(self, mode):
            return getattr(self, mode)() if mode != None else self.default

        def DRAW():
            return Shape.edge_color.default

        def EXPAND():
            return Shape.edge_color.default

        def OFFSET():
            return Shape.edge_color.default

        def GRAB():
            return Shape.edge_color.default

        def ROTATE():
            return Shape.edge_color.default

        def BEVEL():
            return (*preference().BC_bluebox_color[0:3], 1.0)

        def SOLIDIFY():
            return (*preference().BC_yellowbox_color[0:3], 1.0)

        def ARRAY():
            return (*preference().BC_graybox_color[0:3], 1.0)

    class widget_color:
        default = (0.1, 0.1, 0.1, 0.95)
        faded = (0.1, 0.1, 0.1, 0.4)

        def __new__(self, mode):
            return getattr(self, mode)()

        def DRAW():
            return Shape.widget_color.default

        def EXPAND():
            return Shape.widget_color.default

        def OFFSET():
            return Shape.widget_color.default
