import bpy
from bpy.props import *
from ... utils.objects import set_active, current_selected_status, mesh_of_activeobj_select
from ... utils.modifiers import apply_modifier
from ... preferences import preference, BoxCutter_enable_hops
from ... material import assign_material
from . bool_utils import hide_for_cycles_preview


class BoxCutterBoolUnion(bpy.types.Operator):
    bl_idname = "boxcutter.bool_union"
    bl_label = "Box Cutter Union Boolean"
    bl_options = {'REGISTER', 'UNDO'}

    remove_cutters = BoolProperty(name="Remove Cutters",
                                  description="Remove Cutters",
                                  default=True)

    apply_cutters = BoolProperty(name="Apply Cutters",
                                 description="Apply Cutters",
                                 default=True)

    def execute(self, context):
        if self.apply_cutters:
            use_bmesh_boolean(context, self.remove_cutters, 'UNION')
        else:
            use_bmesh_mod_boolean(context, 'UNION')

        return {'FINISHED'}


class BoxCutterBoolDifference(bpy.types.Operator):
    bl_idname = "boxcutter.bool_difference"
    bl_label = "Box Cutter Difference Boolean"
    bl_options = {'REGISTER', 'UNDO'}

    remove_cutters = BoolProperty(name="Remove Cutters",
                                  description="Remove Cutters",
                                  default=True)

    apply_cutters = BoolProperty(name="Apply Cutters",
                                 description="Apply Cutters",
                                 default=True)

    def execute(self, context):
        if self.apply_cutters:
            use_multi_boolean(context, self.remove_cutters, 'DIFFERENCE')
        else:
            use_multi_boolean(context, self.remove_cutters, 'DIFFERENCE', False)

        return {'FINISHED'}


class BoxCutterBoolIntersect(bpy.types.Operator):
    bl_idname = "boxcutter.bool_intersect"
    bl_label = "Box Cutter Intersect Boolean"
    bl_options = {'REGISTER', 'UNDO'}

    remove_cutters = BoolProperty(name="Remove Cutters",
                                  description="Remove Cutters",
                                  default=True)

    apply_cutters = BoolProperty(name="Apply Cutters",
                                 description="Apply Cutters",
                                 default=True)

    def execute(self, context):
        if self.apply_cutters:
            use_carve_boolean(context, self.remove_cutters, 'INTERSECT')
        else:
            use_bmesh_mod_boolean(context, 'INTERSECT')

        return {'FINISHED'}


def bc_diff_boolean(context, in_editmode, apply_cutters, remove_cutters, boolean_method='DIFFERENCE'):
    if apply_cutters:
        use_multi_boolean(context, in_editmode, remove_cutters, boolean_method, True)
    else:
        use_multi_boolean(context, in_editmode, remove_cutters, boolean_method, False)


def use_bmesh_boolean(context, remove_cutters, boolean_method='DIFFERENCE'):
    active_object, other_objects, other_object = current_selected_status()
    if not other_objects:
        assign_material(context, other_object)
    for obj in other_objects:
        if BoxCutter_enable_hops():
            obj.hops.status = "BOOLSHAPE"
        assign_material(context, obj)
        obj.hide_render = True
        obj.draw_type = 'WIRE'
        boolean = active_object.modifiers.new("Boolean", "BOOLEAN")
        if boolean_method == 'DIFFERENCE':
            boolean.operation = 'DIFFERENCE'
        elif boolean_method == 'UNION':
            boolean.operation = 'UNION'
        elif boolean_method == 'INTERSECT':
            boolean.operation = 'INTERSECT'
        boolean.object = obj
    apply_modifier(boolean)
    bpy.ops.object.select_all(action='DESELECT')
    if remove_cutters:
        obj.select = True
        bpy.ops.object.delete(use_global=False)
    else:
        obj.cycles_visibility.camera = False
        obj.cycles_visibility.diffuse = False
        obj.cycles_visibility.glossy = False
        obj.cycles_visibility.transmission = False
        obj.cycles_visibility.scatter = False
        obj.cycles_visibility.shadow = False


def use_bmesh_mod_boolean(context, boolean_method='DIFFERENCE'):
    active_object, other_objects, other_object = current_selected_status()
    if not other_objects:
        assign_material(context, other_object)

    for obj in other_objects:
        if BoxCutter_enable_hops():
            obj.hops.status = "BOOLSHAPE"
        assign_material(context, obj)
        obj.hide_render = True
        obj.draw_type = 'WIRE'
        hide_for_cycles_preview(context, obj)

        boolean = active_object.modifiers.new("Boolean", "BOOLEAN")
        if boolean_method == 'DIFFERENCE':
            boolean.operation = 'DIFFERENCE'
        elif boolean_method == 'UNION':
            boolean.operation = 'UNION'
        elif boolean_method == 'INTERSECT':
            boolean.operation = 'INTERSECT'
        boolean.object = obj
    set_active(active_object, select=True, only_select=True)


def use_multi_boolean(context, in_editmode, remove_cutters, boolean_method='DIFFERENCE', apply_mod=True):
    active_object, other_objects, other_object = current_selected_status()

    if in_editmode:
        bpy.ops.object.mode_set(mode='OBJECT')

    if preference().workflow == "NONDESTRUCTIVE":
        for f in active_object.data.polygons:
            f.use_smooth = True
        for obj in other_objects:
            obj.data.use_auto_smooth = True
            for mod in obj.modifiers:
                if mod.type == "BEVEL":
                    mod.limit_method = 'ANGLE'

    for obj in other_objects:
        bpy.ops.object.select_all(action='DESELECT')
        obj.select = True
        active_object.select = True
        bpy.context.scene.objects.active = obj

        assign_material(context, active_object)

        active_object.hide_render = True
        active_object.draw_type = 'WIRE'
        boolean = obj.modifiers.new("BC_Boolean", "BOOLEAN")
        if boolean_method == 'DIFFERENCE':
            boolean.operation = 'DIFFERENCE'
        elif boolean_method == 'UNION':
            boolean.operation = 'UNION'
        elif boolean_method == 'INTERSECT':
            boolean.operation = 'INTERSECT'
        boolean.object = active_object

        if preference().workflow == "NONDESTRUCTIVE":
            for modifier in reversed(modifiers_by_name(obj)):
                while obj.modifiers.find(modifier.name) != 0:
                    bpy.ops.object.modifier_move_up(modifier=modifier.name)

        # boolean.name = "BC_Boolean"

        if apply_mod:
            apply_modifier(boolean)

    if BoxCutter_enable_hops():
        if in_editmode:
            bpy.ops.hops.soft_sharpen()
        active_object.hops.status = "BOOLSHAPE"

    bpy.ops.object.select_all(action='DESELECT')

    active_object.cycles_visibility.camera = False
    active_object.cycles_visibility.diffuse = False
    active_object.cycles_visibility.glossy = False
    active_object.cycles_visibility.transmission = False
    active_object.cycles_visibility.scatter = False
    active_object.cycles_visibility.shadow = False

    if apply_mod:
        if remove_cutters:
            active_object.select = True
            bpy.ops.object.delete(use_global=False)

    if in_editmode:
        object = bpy.context.active_object
        bpy.ops.object.select_all(action='DESELECT')
        object.select = True
        bpy.ops.object.mode_set(mode='EDIT')


def modifiers_by_name(obj):
    return [x for x in obj.modifiers if x.type in {"BOOLEAN", "MIRROR", "SOLIDIFY"}]
