import bpy
from bgl import *
from bpy.props import *
from ... material import assign_material
from ... utils.modifiers import apply_modifier
from ... preferences import preference, BoxCutter_enable_hops
from . bool_utils import hide_for_cycles_preview


class BoxCutterSlash(bpy.types.Operator):
    bl_idname = "boxcutter.slash"
    bl_label = "Slash"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Splits the primary mesh using the secondary mesh"

    remove_cutters = BoolProperty(name="Remove Cutters",
                                  description="Remove Cutters",
                                  default=False)

    prevent_moving = BoolProperty(name="Prevent From moving",
                                  description="Prevent From moving",
                                  default=False)

    @classmethod
    def poll(cls, context):
        object = context.active_object
        if object is None: return False
        return object.type == "MESH" and object.mode == "OBJECT"

    def draw(self, context):
        layout = self.layout
        box = layout.box()
        col = box.column()
        col.prop(self, "prevent_moving")
        if preference().workflow == "DESTRUCTIVE":
            col.prop(self, "remove_cutters")
        else:
            col.label(text="")
        col.separator()
        # col.label(text="Slash Parameters")
        colrow = col.row(align=True)
        colrow.prop(preference(), "workflow", text="Grid Type", expand=True)

    def execute(self, context):
        slash(self, context, self.remove_cutters, self.prevent_moving)
        return {'FINISHED'}


def slash(self, context, remove_cutters, stay_in_edit_mode, first_selected_objects=None):

    cutters = cutter_objects()
    cutters = [n for n in cutters if n not in first_selected_objects]
    prevent_moving = False

    if stay_in_edit_mode:
        bpy.ops.object.mode_set(mode='OBJECT')
        slash_obj(self, context, remove_cutters, prevent_moving, cutters, first_selected_objects=None)
        if remove_cutters:
            for cutter in cutters:
                context.scene.objects.unlink(cutter)
                bpy.data.objects.remove(cutter)
        bpy.ops.object.mode_set(mode='EDIT')

    else:
        for first_obj in first_selected_objects:
            bpy.ops.object.select_all(action='DESELECT')
            first_obj.select = True
            bpy.context.scene.objects.active = first_obj

            slash_obj(self, context, remove_cutters, prevent_moving, cutters, first_selected_objects=None)

        bpy.ops.object.select_all(action='DESELECT')
        if remove_cutters:
            for cutter in cutters:
                context.scene.objects.unlink(cutter)
                bpy.data.objects.remove(cutter)
            bpy.ops.object.delete(use_global=False)
        for obj in first_selected_objects:
            obj.select = True
            bpy.context.scene.objects.active = first_obj


def slash_obj(self, context, remove_cutters, prevent_moving, cutters, first_selected_objects):
    # cutters = get_cutter_objects(self, context)
    object = context.active_object
    cutter = cutters[0]
    scene = context.scene

    for f in cutter.data.polygons:
        f.use_smooth = True

    new_obj = object.copy()
    new_obj.data = object.data.copy()
    # new_obj.animation_data_clear()
    scene.objects.link(new_obj)

    if BoxCutter_enable_hops():
        cutter.hops.status = "BOOLSHAPE"
    cutter.hide_render = True
    cutter.draw_type = "WIRE"

    modifier_active_obj = object.modifiers.new(name="BC_Boolean", type="BOOLEAN")
    modifier_active_obj.operation = "DIFFERENCE"
    modifier_active_obj.object = cutter

    if preference().workflow == "NONDESTRUCTIVE":
        for modifier in reversed(modifiers_by_name(object)):
            while object.modifiers.find(modifier.name) != 0:
                bpy.ops.object.modifier_move_up(modifier=modifier.name)

    modifier_new_obj = new_obj.modifiers.new(name="BC_Boolean", type="BOOLEAN")
    modifier_new_obj.operation = "INTERSECT"
    modifier_new_obj.object = cutter

    if preference().workflow == "NONDESTRUCTIVE":
        context.scene.objects.active = new_obj
        new_obj.select = True
        # set_active(new_obj, select=True, only_select=True)
        for modifier in reversed(modifiers_by_name(new_obj)):
            while new_obj.modifiers.find(modifier.name) != 0:
                bpy.ops.object.modifier_move_up(modifier=modifier.name)

    if prevent_moving:
        object.lock_location[0] = True
        object.lock_location[1] = True
        object.lock_location[2] = True
        object.lock_rotation[0] = True
        object.lock_rotation[1] = True
        object.lock_rotation[2] = True
        object.lock_scale[0] = True
        object.lock_scale[1] = True
        object.lock_scale[2] = True
        new_obj.lock_location[0] = True
        new_obj.lock_location[1] = True
        new_obj.lock_location[2] = True
        new_obj.lock_rotation[0] = True
        new_obj.lock_rotation[1] = True
        new_obj.lock_rotation[2] = True
        new_obj.lock_scale[0] = True
        new_obj.lock_scale[1] = True
        new_obj.lock_scale[2] = True

    if preference().workflow == "DESTRUCTIVE":
        for mod in object.modifiers:
            if mod == modifier_active_obj:
                apply_modifier(modifier_active_obj)

        for mod in new_obj.modifiers:
            if mod == modifier_new_obj:
                apply_modifier(modifier_new_obj)

        cutter.select = False
        if BoxCutter_enable_hops():
            bpy.ops.hops.soft_sharpen()

    if preference().workflow == "NONDESTRUCTIVE":
        for mod in object.modifiers:
            if mod.type == "BEVEL":
                mod.limit_method = 'ANGLE'

        for mod in new_obj.modifiers:
            if mod.type == "BEVEL":
                mod.limit_method = 'ANGLE'

    assign_material(context, new_obj, replace=True)


def get_cutter_objects(self, context):
    selection = context.selected_objects
    active = context.active_object
    return [object for object in selection if object != active and object.type == "MESH"]


def modifiers_by_name(obj):
    return [x for x in obj.modifiers if x.type in {"BOOLEAN", "MIRROR", "SOLIDIFY"}]


class BoxcutterSplitcheck(bpy.types.Operator):
    bl_idname = "boxcutter.slice_check_old"
    bl_label = "Complex Split Check"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = ""

    remove_cutters = BoolProperty(name="Remove Cutters",
                                  description="Remove Cutters",
                                  default=False)

    stay_in_edit_mode = BoolProperty(name="Remove Cutters",
                                     description="Remove Cutters",
                                     default=False)

    @classmethod
    def poll(cls, context):
        object = context.active_object
        if object is None: return False
        return object.type == "MESH" and object.mode == "OBJECT"

    def execute(self, context):
        slash2(self, context, self.remove_cutters, self.stay_in_edit_mode)
        return {'FINISHED'}


def slash2(self, context, remove_cutters, stay_in_edit_mode, first_selected_objects=None):

    cutters = cutter_objects()
    cutters = [n for n in cutters if n not in first_selected_objects]

    if stay_in_edit_mode:
        bpy.ops.object.mode_set(mode='OBJECT')
        slash_single_obj(self, context, remove_cutters, stay_in_edit_mode, cutters, first_selected_objects=None)
        bpy.ops.object.mode_set(mode='EDIT')

    else:
        for first_obj in first_selected_objects:
            bpy.ops.object.select_all(action='DESELECT')
            first_obj.select = True
            bpy.context.scene.objects.active = first_obj

            slash_single_obj(self, context, remove_cutters, stay_in_edit_mode, cutters, first_selected_objects=None)

        bpy.ops.object.select_all(action='DESELECT')
        if remove_cutters:
            for cutter in cutters:
                bpy.data.objects.remove(cutter)
            bpy.ops.object.delete(use_global=False)
        for obj in first_selected_objects:
            obj.select = True
            bpy.context.scene.objects.active = first_obj


def slash_single_obj(self, context, remove_cutters, stay_in_edit_mode, cutters, first_selected_objects=None):

    split_mesh = preference().BoxCutter_split_mesh
    # cutters = cutter_objects()
    object = bpy.context.active_object

    for cutter in cutters:
        cutter.select = True
        cutter.draw_type = "WIRE"
        cutter.hide_render = True
        solidify = cutter.modifiers.new(name="temp_solid", type="SOLIDIFY")
        solidify.thickness = 0.0001
        solidify.offset = 1
        modifier = object.modifiers.new(name="temp", type="BOOLEAN")
        modifier.operation = "DIFFERENCE"
        modifier.object = cutter
        for mod in object.modifiers:
            if mod.type == "BOOLEAN":
                bpy.ops.object.modifier_apply(apply_as="DATA", modifier="temp")

        cutter.modifiers.remove(solidify)

    if BoxCutter_enable_hops():
        bpy.ops.hops.soft_sharpen()
    selected = bpy.context.selected_objects
    split_the_mesh(self, context, split_mesh, object, selected)

    all_selected = cutter_objects()
    splited = [x for x in all_selected if x not in cutters]

    bpy.ops.object.select_all(action='DESELECT')
    for obj in splited:
        obj.select = True
        assign_material(context, obj)
        bpy.context.scene.objects.active = obj
        if preference().BoxCutter_enable_hops is True:
            if obj.hops.status == "CSTEP":
                me = obj.data
                for e in me.edges:
                    e.select = False
                    if e.bevel_weight > 0:
                        e.select = True
                bpy.ops.object.mode_set(mode='EDIT')
                bpy.ops.mesh.hide(unselected=True)
                bpy.ops.object.mode_set(mode='OBJECT')

    bpy.ops.object.select_all(action='DESELECT')
    for obj in cutters:
        obj.select = True
        if BoxCutter_enable_hops():
            obj.hops.status = "BOOLSHAPE"
        obj.hide_render = True
        obj.draw_type = 'WIRE'
        hide_for_cycles_preview(context, obj)

    bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY')
    if stay_in_edit_mode:
        if remove_cutters:
            for cutter in cutters:
                bpy.data.objects.remove(cutter)

    bpy.ops.object.select_all(action='DESELECT')
    bpy.context.scene.objects.active = object
    object.select = True

    if preference().BoxCutter_enable_hops is True:

        if not split_mesh:
            bpy.ops.object.mode_set(mode='EDIT')
            bpy.ops.mesh.select_all(action='DESELECT')
            bpy.ops.object.mode_set(mode='OBJECT')

        if object.hops.status == "CSTEP":
            me = object.data
            for e in me.edges:
                if e.bevel_weight > 0:
                    e.select = True
            bpy.ops.object.mode_set(mode='EDIT')
            bpy.ops.mesh.hide(unselected=True)
            bpy.ops.object.mode_set(mode='OBJECT')


# @staticmethod
def cutter_objects():
    selection = bpy.context.selected_objects
    active = bpy.context.active_object
    return [object for object in selection if object != active and object.type == "MESH"]


def select_inside(self, context, active, selected):
    # selected = bpy.context.selected_objects
    # active = bpy.context.active_object
    selected.remove(active)
    max_dist = 1.84467e+19

    bpy.ops.object.mode_set(mode='EDIT')
    bpy.ops.mesh.reveal()
    bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='VERT')
    bpy.ops.mesh.select_all(action='DESELECT')
    bpy.ops.object.mode_set(mode='OBJECT')

    active.select = False
    bpy.ops.object.transform_apply(location=True, rotation=True, scale=True)
    active.select = True

    for obj in selected:
        for v in active.data.vertices:
            if is_inside((active.matrix_world * v.co), max_dist, obj):
                v.select = True

    bpy.ops.object.mode_set(mode='EDIT')
    bpy.ops.mesh.select_linked(delimit=set())
    bpy.ops.mesh.select_mode(use_extend=False, use_expand=False, type='EDGE')
    bpy.ops.object.mode_set(mode='OBJECT')


def is_inside(p, max_dist, obj):
    hit, point, normal, face = obj.closest_point_on_mesh(p, max_dist)
    p2 = (obj.matrix_world*point-p)
    v = p2.dot(normal)
    return not(v < 0.0)


def split_the_mesh(self, context, split_mesh,  active, selected):
    select_inside(self, context,  active, selected)
    if split_mesh:
        bpy.ops.object.mode_set(mode='EDIT')
        try:
            bpy.ops.mesh.separate()
        except:
            pass
        bpy.ops.object.mode_set(mode='OBJECT')
