import time

from copy import deepcopy as copy

import bpy

from bpy.props import EnumProperty

from mathutils import Vector, Matrix

from .. boolean_operators.bool_bmesh import bc_diff_boolean
from .. boolean_operators.split_check import slash

from ... graphic.draw_operators.shape import Shape

from ... utils.modifiers import live
from ... utils.space_3d import location3d_to_point2d, ray_cast
from ... utils.view_3d import view_orientation

from ... preferences import preference


class BoxCutterDrawShape(bpy.types.Operator):
    bl_idname = 'boxcutter.draw_shape'
    bl_label = 'Box Cutter Draw Cube'
    bl_options = {'REGISTER', 'UNDO', 'GRAB_CURSOR'}

    mouse = Vector()
    mouse_start = Vector()

    view_matrix = Matrix()
    view_location = Vector()
    view_distance = float()

    lmb_click_time = float()

    initial_selected_objects = list()

    mode = EnumProperty(
        name = 'Mode',
        description = 'Cutter mode',
        items = [
            ('CUT', 'Cut', ''),
            ('SLICE', 'Slice', ''),
            ('MAKE', 'Make', ''),
            ('JOIN', 'Join', '')],
        default = 'CUT')

    shape_mode = 'DRAW' # EXPAND, OFFSET, None, BEVEL, GRAB, ROTATE, SOLIDIFY, ARRAY

    axis_mode = 'X'
    axis_index = int()
    constrain = bool()

    auto_expand = True

    #TODO: origin_mode = 'CORNER' # CENTER
    #TODO: constrain_mode = 'LOCAL' # SURFACE, VIEW, TRACKBALL

    origin = Vector()
    modified = bool()
    start_matrix = Matrix()

    use_depth = bool()
    remove_cutters = True

    active_object = None

    ray_object = None
    hit_location = Vector()
    face_normal = Vector()
    depth_normal = Vector()
    face_index = int()
    face_center = Vector()

    shape = None
    update = bool()

    lat = None
    latmod = None

    start_mod = tuple()

    bevel_mode = 'ROLL' # ROUND TODO: YAW, PITCH
    #TODO: bevel_mode_type = 'HALF' # FULL, QUARTER
    bevmod = None
    bevmod_use_clamp_overlap = True
    bevmod_start_width = float()

    mirmod = None

    solmod = None
    solmod_start_thickness = float()

    arrmod = None
    arrmod_count = 1
    arrmod_start_offset = 1

    empty = None

    angle = float()
    angle_track = float()
    angle_text = str()

    active_widget = str()
    widget_location = {
        'DRAW': Vector(),
        'EXPAND': Vector(),
        'OFFSET': Vector()}

    widget_color = {
        'DRAW': tuple(),
        'EXPAND': tuple(),
        'OFFSET': tuple()}

    draw_dot_point = int()
    expand_point = int()

    draw_cutoff = bool()

    #TODO: create collection props to store these and parse the info (shape feed/library system)
    verts = [Vector((-0.5, -0.5, -0.5)), Vector((-0.5, -0.5,  0.5)),
             Vector((-0.5,  0.5, -0.5)), Vector((-0.5,  0.5,  0.5)),
             Vector(( 0.5, -0.5, -0.5)), Vector(( 0.5, -0.5,  0.5)),
             Vector(( 0.5,  0.5, -0.5)), Vector(( 0.5,  0.5,  0.5))]

    edges = [(2, 0), (0, 1), (1, 3), (3, 2),
             (6, 2), (3, 7), (7, 6), (4, 6),
             (7, 5), (5, 4), (0, 4), (5, 1)]

    faces = [(0, 1, 3, 2), (2, 3, 7, 6), (6, 7, 5, 4),
             (4, 5, 1, 0), (2, 6, 4, 0), (7, 3, 1, 5)]

    numbers = [
        'ZERO', 'ONE', 'TWO', 'THREE', 'FOUR',
        'FIVE', 'SIX', 'SEVEN', 'EIGHT', 'NINE']

    numpad_numbers = [
        'NUMPAD_0', 'NUMPAD_1', 'NUMPAD_2', 'NUMPAD_3', 'NUMPAD_4',
        'NUMPAD_5', 'NUMPAD_6', 'NUMPAD_7', 'NUMPAD_8', 'NUMPAD_9']

    def draw(self, context):
        layout = self.layout
        box = layout.box()
        col = box.column()
        col.label('Box Cutter')

    def invoke(self, context, event):
        preference().BoxCutter_help_item = 'SHAPE'

        removeable = []
        for live_name in live:
            shape = False
            for object in bpy.data.objects:
                if live_name in bpy.data.objects and object == bpy.data.objects[live_name]:
                    shape = True
            if not shape:
                removeable.append(live_name)

        for live_name in removeable:
            live.remove(live_name)

        if preference().workflow == "NONDESTRUCTIVE":
            self.remove_cutters = False
        else:
            self.remove_cutters = True

        use_2d = not context.space_data.region_3d.is_perspective and view_orientation(context.space_data.region_3d.view_rotation) != 'UNDEFINED'

        self.initial_selected_objects = bpy.context.selected_objects

        is_local = bool(context.space_data.local_view)

        if is_local:
            self.view_matrix = context.region_data.view_matrix
            self.view_location = Vector(context.region_data.view_location[:])
            self.view_distance = context.region_data.view_distance

            bpy.ops.view3d.localview()

        self.ray_object, self.hit_location, self.face_normal, self.face_index, self.face_center = ray_cast(self, context, event)

        if not use_2d and self.ray_object:
            self.mouse = Vector((event.mouse_region_x, event.mouse_region_y))
            self.lmb_click_time = time.clock()

            self.org_wire_state = self.active_object.show_wire
            self.active_object.show_wire = True
            self.active_object.show_all_edges = True

            self.update = True
            Shape(self, context)

            if is_local:
                self.shape.hide = False
                self.lat.hide = False

                bpy.ops.view3d.localview()

                self.shape.hide = True
                self.lat.hide = True

                context.region_data.view_matrix = self.view_matrix
                context.region_data.view_location = self.view_location
                context.region_data.view_distance = self.view_distance

            self.ray_object.select = False
            self.shape.select = False
            self.lat.select = False

            self.handler2d = bpy.types.SpaceView3D.draw_handler_add(Shape.draw2d, (self, context), 'WINDOW', 'POST_PIXEL')
            self.handler3d = bpy.types.SpaceView3D.draw_handler_add(Shape.draw3d, (self, context), 'WINDOW', 'POST_VIEW')
            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}

        if self.ray_object:
            self.cleanup(ray_object=True)

        if is_local:
            bpy.ops.view3d.localview()

            context.region_data.view_matrix = self.view_matrix
            context.region_data.view_location = self.view_location
            context.region_data.view_distance = self.view_distance

        if len(bpy.context.selected_objects) > 0:
            if preference().BoxCutter_enable_borders and len(bpy.context.selected_objects) > 1:
                if event.ctrl:
                    bpy.ops.boxcutter.draw_border('INVOKE_DEFAULT')
                    return {'RUNNING_MODAL'}
            else:
                if event.ctrl:
                    if event.shift:
                        bpy.ops.boxcutter.draw_poly('INVOKE_DEFAULT', mode='SLICE')
                        return {'RUNNING_MODAL'}
                    else:
                        bpy.ops.boxcutter.draw_poly('INVOKE_DEFAULT', mode='CUT')
                        return {'RUNNING_MODAL'}

        elif len(bpy.context.selected_objects) == 0:
            if event.ctrl:
                if event.shift:
                    bpy.ops.boxcutter.add_cube()
                    return {'RUNNING_MODAL'}
                else:
                    bpy.ops.boxcutter.draw_poly('INVOKE_DEFAULT', mode='MAKE')
                    return {'RUNNING_MODAL'}

        return {'CANCELLED'}

    def modal(self, context, event):
        double_click_time = context.user_preferences.inputs.mouse_double_click_time * 0.001
        self.mouse = Vector((event.mouse_region_x, event.mouse_region_y))

        if event.type in {'ESC', 'RIGHTMOUSE'} and event.value == 'RELEASE':
            if not self.shape_mode or not self.modified:
                self.exit(context)
                return {'CANCELLED'}
            else:
                self.exit_mode()

        elif event.type in {'RET', 'NUMPAD_ENTER'}:
            if event.alt: self.use_depth = True
            if event.shift:
                self.remove_cutters = False

            self.execute(context)
            return {'FINISHED'}

        elif event.type == 'LEFTMOUSE':
            if event.value == 'PRESS':
                if self.active_widget:
                    self.enter_mode(self.active_widget, mouse=True, matrix=True, modified=True)

                else:
                    if self.shape_mode is None:
                        if (time.clock() - self.lmb_click_time) <= double_click_time:
                            if event.alt: self.use_depth = True
                            if event.shift:
                                self.remove_cutters = False

                            self.execute(context)
                            return {'FINISHED'}

                        else:
                            self.enter_mode('BEVEL', mouse=True, modified=True)
                            self.lmb_click_time = time.clock()

                    elif self.shape_mode == 'EXPAND' and not self.modified:
                        if event.alt: self.use_depth = True
                        if event.shift:
                            self.remove_cutters = False

                        self.execute(context)
                        return {'FINISHED'}

                    else:
                        self.enter_mode(None, mouse=True, modified=True)

            elif event.value == 'RELEASE':
                if self.shape_mode == 'DRAW':
                    if self.auto_expand:
                        self.enter_mode('EXPAND')
                        self.auto_expand = False

                    else:
                        self.enter_mode(None, modified=True)

                else:
                    self.enter_mode(None, modified=True)

                    if self.bevmod:
                        self.enter_mode(None, bevmod_width=True, modified=True)

                    if self.arrmod:
                        self.arrmod_start_count = self.arrmod.count

        elif event.type in {'MIDDLEMOUSE'}:
            if self.shape_mode != 'ROTATE':
                return {'PASS_THROUGH'}

        elif event.type in {'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
            if self.shape_mode not in {'BEVEL', 'ARRAY'}:
                return {'PASS_THROUGH'}

            elif event.type == 'WHEELUPMOUSE':
                if self.shape_mode == 'BEVEL':
                    if self.bevmod:
                        self.update = True
                        self.bevmod.segments += 1

                elif self.shape_mode == 'ARRAY':
                    if self.arrmod:
                        self.update = True
                        self.arrmod.count += 1

            elif event.type == 'WHEELDOWNMOUSE':
                if self.shape_mode == 'BEVEL':
                    if self.bevmod:
                        self.update = True
                        self.bevmod.segments -= 1

                elif self.shape_mode == 'ARRAY':
                    if self.arrmod:
                        self.update = True
                        self.arrmod.count -= 1

        elif event.type in {'EQUAL', 'NUMPAD_PLUS'} and event.value == 'PRESS':
            if self.bevmod:
                self.update = True
                self.bevmod.segments += 1

            elif self.shape_mode == 'ARRAY':
                self.update = True
                self.arrmod.count += 1

        elif event.type in {'MINUS', 'NUMPAD_MINUS'} and event.value == 'PRESS':
            if self.bevmod:
                self.update = True
                self.bevmod.segments -= 1

            elif self.shape_mode == 'ARRAY':
                self.update = True
                self.arrmod.count -= 1

        elif event.type == 'Q' and event.value == 'RELEASE':
            if event.alt:
                self.bevmod_use_clamp_overlap = not self.bevmod_use_clamp_overlap

            else:
                if self.bevmod:
                    self.update = True
                    if self.bevel_mode == 'ROLL':
                        self.bevel_mode = 'ROUND'
                        for edge in (0, 1, 2, 3, 4, 5, 10, 11):
                            self.shape.data.edges[edge].bevel_weight = 1
                        self.shape.data.validate()

                        if self.shape.dimensions.z < self.bevmod.width:
                            self.bevmod.width = self.shape.dimensions.z - 0.00001

                    else:
                        self.bevel_mode = 'ROLL'
                        for edge in (0, 1, 2, 3):
                            self.shape.data.edges[edge].bevel_weight = 0
                        self.shape.data.validate()

        elif event.type == 'D' and event.value == 'RELEASE':
            self.enter_mode('DRAW', modified=True)
            self.constrain = False

        elif event.type == 'E' and event.value == 'RELEASE':
            self.enter_mode('EXPAND', modified=True)

        elif event.type == 'O' and event.value == 'RELEASE':
            self.enter_mode('OFFSET', modified=True)

        elif event.type in {'G', 'SPACE'} and event.value == 'RELEASE':
            self.enter_mode('GRAB', modified=True)
            self.constrain = False

            self.start_matrix = self.shape.matrix_world.copy()
            self.origin = self.start_matrix.translation

        elif event.type == 'R' and event.value == 'RELEASE':
            self.enter_mode('ROTATE', matrix=True, modified=True)
            self.constrain = False

        elif event.type == 'B' and event.value == 'RELEASE':
            self.enter_mode('BEVEL', mouse=True, bevmod_width=True, modified=True)

        elif event.type == 'T' and event.value == 'RELEASE':
            self.enter_mode('SOLIDIFY', mouse=True, solmod_thickness=True, modified=True)

        elif event.type == 'V' or event.type in {*self.numbers, *self.numpad_numbers} and event.alt and event.value == 'RELEASE':
            self.enter_mode('ARRAY', mouse=True, arrmod_offset=True, modified=True)

            if event.type != 'V':
                if event.type[-1].isdigit():
                    self.arrmod_count = int(event.type[-1])
                else:
                    self.arrmod_count = self.numbers.index(event.type)

        elif event.type == 'W' and event.value == 'RELEASE':
            self.active_object.show_wire = not self.active_object.show_wire
            self.active_object.show_all_edges = not self.active_object.show_all_edges

            self.shape.show_wire = not self.shape.show_wire
            self.shape.show_all_edges = not self.shape.show_all_edges

        elif self.shape_mode in {'DRAW', 'GRAB', 'ROTATE', 'ARRAY'} and event.type in {'X', 'Y', 'Z'} and event.value == 'RELEASE':
            if self.shape_mode in {'GRAB', 'ROTATE'} or self.shape_mode == 'DRAW' and event.type != 'Z':
                self.constrain = True

                if self.shape_mode != 'DRAW':
                    self.shape.matrix_world = self.start_matrix
                    self.lat.matrix_world = self.start_matrix

                    self.angle = 0

            self.axis_mode = event.type

            if self.shape_mode == 'ARRAY':
                mode = [axis == self.axis_mode for axis in 'XYZ']
                self.axis_index = mode.index(True)

        elif event.type == 'H' and event.value == 'RELEASE':
            if not event.ctrl and not event.alt:
                if not self.draw_cutoff:
                    Shape.hide = not Shape.hide
                    self.lat.hide = not Shape.hide
                    self.shape.hide = not Shape.hide

            else:
                if event.ctrl:
                    Shape.display_negative = not Shape.display_negative
                if event.alt:
                    Shape.display_bbox = not Shape.display_bbox

        elif event.type == 'A' and event.value == 'RELEASE':
            self.mode = 'MAKE' if self.mode != 'MAKE' else 'CUT'

        elif event.type == 'X' and event.value == 'RELEASE':
            self.mode = 'SLICE' if self.mode != 'SLICE' else 'CUT'

        elif event.type == 'Z' and not self.shape_mode in {'GRAB', 'ROTATE'}:
            if event.value == 'RELEASE':
                if event.shift:
                    preference().BoxCutter_enable_grid = not preference().BoxCutter_enable_grid
                else:
                    self.mode = 'CUT'

        elif event.type == 'J' and event.value == 'RELEASE':
            self.mode = 'JOIN' if self.mode != 'JOIN' else 'CUT'

        elif event.type in {'ONE', 'NUMPAD_1'} and event.value == 'RELEASE':
            self.update = True
            preference().BoxCutter_mirror_axis_x = not preference().BoxCutter_mirror_axis_x

            Shape.update.mirror(self, context, event)

        elif event.type in {'TWO', 'NUMPAD_2'} and event.value == 'RELEASE':
            self.update = True
            preference().BoxCutter_mirror_axis_y = not preference().BoxCutter_mirror_axis_y

            Shape.update.mirror(self, context, event)

        elif event.type in {'THREE', 'NUMPAD_3'} and event.value == 'RELEASE':
            #TODO: add mirror axis z preference to d menu
            self.update = True
            preference().BoxCutter_mirror_axis_z = not preference().BoxCutter_mirror_axis_z

            Shape.update.mirror(self, context, event)

        elif event.type == 'TAB' and event.value == 'RELEASE':
            preference().BoxCutter_help_item = 'SHAPE_EDIT'
            self.enter_mode(None, modified=True)

        Shape.update(self, context, event)
        context.area.tag_redraw()
        return {'RUNNING_MODAL'}

    def execute(self, context):
        self.shape.select = True
        self.shape.hide = False
        context.scene.objects.active = self.shape

        if self.shape.dimensions.z < 0.005:
            dimensions = [abs(self.ray_object.dimensions.x), abs(self.ray_object.dimensions.y), abs(self.ray_object.dimensions.z)]
            thickness = preference().BoxCutter_cut_depth if self.use_depth else max(dimensions) * 1.5

            for point in (0, 2, 4, 6):
                self.lat.data.points[point].co_deform.z -= thickness

        if self.mode == 'JOIN':
            for point in Shape.points['front']:
                self.lat.data.points[point].co_deform.z -= 0.001

        bpy.ops.object.modifier_apply(apply_as='DATA', modifier=self.latmod.name)

        if self.bevmod:
            if self.bevmod.width < 0.0001:
                self.shape.modifiers.remove(self.bevmod)
                self.bevmod = None

        if self.mirmod:
            if not self.mirmod.use_x and not self.mirmod.use_y and not self.mirmod.use_z:
                self.shape.modifiers.remove(self.mirmod)
                self.mirmod = None

        if not self.remove_cutters:
            if self.shape.name in live:
                self.shape.name += '.{0:03d}'.format(
                    len([name for name in live if name == self.shape.name or name[:-4] == self.shape.name]))

            live.append(self.shape.name)

        apply_cutters = self.remove_cutters

        if apply_cutters:
            if self.arrmod:
                bpy.ops.object.modifier_apply(apply_as='DATA', modifier=self.arrmod.name)
            if self.mirmod:
                bpy.ops.object.modifier_apply(apply_as='DATA', modifier=self.mirmod.name)
            if self.solmod:
                bpy.ops.object.modifier_apply(apply_as='DATA', modifier=self.solmod.name)
            if self.bevmod:
                bpy.ops.object.modifier_apply(apply_as='DATA', modifier=self.bevmod.name)

                for edge in self.shape.data.edges:
                    edge.bevel_weight = 0

        elif self.shape_mode == 'MAKE':
            self.shape.draw_type = 'SOLID'

        else:
            self.shape.cycles_visibility.camera = False
            self.shape.cycles_visibility.diffuse = False
            self.shape.cycles_visibility.glossy = False
            self.shape.cycles_visibility.transmission = False
            self.shape.cycles_visibility.scatter = False
            self.shape.cycles_visibility.shadow = False

        self.shape.data.validate()

        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.mesh.select_all(action='SELECT')
        bpy.ops.mesh.remove_doubles()
        bpy.ops.mesh.normals_make_consistent(inside=False)
        bpy.ops.object.editmode_toggle()

        if self.mode == 'CUT':
            self.active_object.select = True
            bc_diff_boolean(context, False, apply_cutters, self.remove_cutters, boolean_method='DIFFERENCE')
            if preference().BoxCutter_enable_hops:
                for obj in self.initial_selected_objects:
                    obj.select = True
                    if obj.hops.status in ('CSHARP', 'CSTEP'):
                        bpy.ops.hops.soft_sharpen()

        elif self.mode == 'SLICE':
            self.active_object.show_wire = self.org_wire_state
            self.active_object.show_all_edges = self.org_wire_state

            bpy.context.scene.objects.active = self.active_object
            slash(self, context, self.remove_cutters, False, self.initial_selected_objects)

        elif self.mode == 'MAKE':
            self.remove_cutters = False
            self.shape.show_wire = self.org_wire_state
            self.shape.show_all_edges = self.org_wire_state
            self.shape.draw_type = 'SOLID'

        elif self.mode == 'JOIN':
            bc_diff_boolean(context, False, apply_cutters, self.remove_cutters, boolean_method='UNION')

        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.mesh.select_all(action='SELECT')
        bpy.ops.mesh.remove_doubles()
        bpy.ops.mesh.normals_make_consistent(inside=False)
        bpy.ops.object.editmode_toggle()

        self.exit(context)
        return {'FINISHED'}

    def enter_mode(self, mode, mouse=False, matrix=False, bevmod_width=False, solmod_thickness=False, arrmod_offset=True, modified=False):
        self.modified = modified
        self.shape_mode = mode

        if mouse:
            self.mouse_start = copy(self.mouse)

        if matrix:
            self.start_matrix = self.shape.matrix_world.copy()

        if bevmod_width:
            self.bevmod_start_width = copy(self.bevmod.width) if self.bevmod else 0

        if solmod_thickness:
            self.solmod_start_thickness = copy(self.solmod.thickness) if self.solmod else 0

        if arrmod_offset:
            if not self.arrmod:
                self.axis_index = 0

            self.arrmod_start_offset = copy(self.arrmod.relative_offset_displace[self.axis_index]) if self.arrmod else 1

        if mode == 'ROTATE':
            self.axis_mode = preference().BC_start_axis
            self.angle = 0

            self.pivot_point = location3d_to_point2d(bpy.context, self.shape.matrix_world.translation)

            self.mouse_start = Vector((self.mouse.x, self.mouse.y)) - self.pivot_point

    def exit_mode(self):

        if self.shape_mode in {'GRAB', 'ROTATE'}:
            self.shape.matrix_world = self.start_matrix
            self.lat.matrix_world = self.start_matrix
            self.start_matrix = Matrix()

        elif self.shape_mode == 'BEVEL':
            if self.bevmod:
                self.bevmod.width = self.bevmod_start_width

        elif self.shape_mode == 'SOLIDIFY':
            if self.solmod:
                self.solmod.thickness = self.solmod_start_thickness

                if abs(self.solmod.thickness) < 0.00001:
                    self.shape.modifiers.remove(self.solmod)
                    self.solmod = None

        elif self.shape_mode == 'ARRAY':
            self.arrmod.relative_offset_displace[self.axis_index] = self.arrmod_start_offset
            self.arrmod.count = self.arrmod_start_count

            if self.arrmod_start_count == 1:
                self.shape.modifiers.remove(self.arrmod)
                self.arrmod = None

        self.update = True
        self.shape_mode = None

    def exit(self, context):
        preference().BoxCutter_help_item = 'MAIN'

        self.active_object.show_wire = self.org_wire_state
        self.active_object.show_all_edges = self.org_wire_state

        self.active_object.select = True
        context.scene.objects.active = self.active_object

        self.cleanup(shape=self.remove_cutters, lattice=True, ray_object=True, empty=self.remove_cutters)

        bpy.types.SpaceView3D.draw_handler_remove(self.handler2d, 'WINDOW')
        bpy.types.SpaceView3D.draw_handler_remove(self.handler3d, 'WINDOW')

    def cleanup(self, shape=False, lattice=False, ray_object=False, empty=False):
        if shape:
            try:
                bpy.data.meshes.remove(self.shape.data)
                bpy.data.objects.remove(self.shape, True)
            except: pass
        if lattice:
            try:
                bpy.data.lattices.remove(self.lat.data)
                bpy.data.objects.remove(self.lat, True)
            except: pass
        if ray_object:
            try:
                bpy.data.meshes.remove(self.ray_object.data)
                bpy.data.objects.remove(self.ray_object, True)
            except: pass
        if empty:
            try: bpy.data.objects.remove(self.empty, True)
            except: pass
