import bpy

from bpy.props import *

from mathutils import Vector, Quaternion
from math import *

from ... graphic.draw_px import draw_view_roll
from ... preferences import preference
from ... utils.view_3d import view_orientation

#TODO: fix exit matrix
class BoxCutterDrawViewRoll(bpy.types.Operator):
    """Draw a screen roll widget"""
    bl_idname = "boxcutter.view_roll"
    bl_label = "Box Cutter View Roll"
    bl_options = {'GRAB_CURSOR'}

    use_snap = BoolProperty(
        name = "Snap",
        description = "Snap during transform",
        default = False
    )

    roll_angle = FloatProperty(
        name = "Angle",
        description = "Angle of the view line",
        subtype = "ANGLE",
        min = 0.0,
        max = 180.0,
        default = 0.0
    )

    #TODO: replace with BC user prefence
    roll_angle_precision = IntProperty(
        name = "Roll angle precision",
        description = "Angle step factor while holding control",
        default = 15
    )

    @classmethod
    def poll(cls, context):
        return context.area.type == 'VIEW_3D' and context.space_data.region_3d.view_perspective != 'CAMERA'


    def invoke(self, context, event):
        self.initial_help_item = preference().BoxCutter_help_item
        preference().BoxCutter_help_item = 'VROLL'

        self.region_3d = context.space_data.region_3d

        self.initial_rotation = self.region_3d.view_rotation.copy()
        self.initial_view_orientation = view_orientation(self.region_3d.view_rotation)

        self.numeric_entry = False
        self.numeric_user_input = ""

        self.is_positive = True
        self.mouse_start = Vector((context.region.width * 0.5, context.region.height * 0.5))
        self.mouse = Vector((event.mouse_region_x, event.mouse_region.y))

        self.roll_factor = 0
        self.roll_angle = self.roll_angle(context)

        if not self.region_3d.is_perspective and self.initial_view_orientation != 'UNDEFINED':
            bpy.ops.view3d.view_orbit(angle=0, type="ORBITLEFT")

        self.draw_handler = bpy.types.SpaceView3D.draw_handler_add(draw_view_roll, (self, context), "WINDOW", "POST_PIXEL")
        context.window_manager.modal_handler_add(self)

        return {'RUNNING_MODAL'}


    def execute(self, context):
        self.exit()


    def modal(self, context, event):
        if 'NUMPAD' in event.type:
            self.numeric_entry = True
            self.is_positive = True

        if self.numeric_entry:
            numpad_type = event.type.split("_")[-1]

            if numpad_type == 'MINUS' and event.value == 'RELEASE':
                if self.numeric_user_input:
                    if self.numeric_user_input[0] != "-":
                        self.numeric_user_input = "-" + self.numeric_user_input

                    else:
                        self.numeric_user_input = self.numeric_user_input[1:]

                else:
                    self.numeric_user_input = "-"

            elif numpad_type == 'PERIOD' and event.value == 'RELEASE':
                if not self.numeric_user_input:
                    self.numeric_user_input = "0."

                elif "." not in self.numeric_user_input:
                    self.numeric_user_input += "."

            elif numpad_type in {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9'} and event.value == "RELEASE":
                self.numeric_user_input += numpad_type

            elif event.type=='BACK_SPACE' and event.value == "RELEASE":
                self.numeric_user_input = self.numeric_user_input[:-1]

            elif event.type in {'SPACE', 'LEFTMOUSE', 'RET', 'NUMPAD_ENTER'}:
                try:
                    self.region_3d.view_rotation = self.initial_rotation * Quaternion(Vector((0, 0, -1)), radians(float(self.numeric_user_input)))

                except:
                    self.cancel(context)

                    return {'CANCELLED'}

                self.execute(context)

                return {'FINISHED'}

            elif event.type in {'ESC', 'RIGHTMOUSE'}:
                self.cancel(context)

                return {'CANCELLED'}

        else:
            self.use_snap = True if event.ctrl else False
            self.use_precision = True if event.shift else False

            if event.type == 'MOUSEMOVE':
                self.mouse = Vector((event.mouse_region_x, event.mouse_region_y))

                self.is_positive = True if self.mouse.y < self.mouse_start.y else False

                if self.use_snap and self.use_precision:
                    self.roll_factor = 1

                elif self.use_snap:
                    self.roll_factor = self.roll_angle_precision

                else:
                    self.roll_factor = None

                self.roll_angle = self.roll_angle(context)
                self.roll_view(context)

            if event.type == 'WHEELUPMOUSE':
                if self.roll_angle_precision < 90:
                    self.roll_angle_precision += 5

            elif event.type == 'WHEELDOWNMOUSE':
                if self.roll_angle_precision > 5:
                    self.roll_angle_precision -= 5

            elif event.type in {'SPACE', 'LEFTMOUSE', 'RET', 'NUMPAD_ENTER'}:
                self.execute(context)

                return {'FINISHED'}

            elif event.type in {'ESC', 'RIGHTMOUSE', 'BACK_SPACE'}:
                self.cancel(context)

                return {'CANCELLED'}

        context.area.tag_redraw()

        return {'RUNNING_MODAL'}


    def cancel(self, context):
        self.region_3d.view_rotation = self.initial_rotation

        if self.initial_view_orientation != 'UNDEFINED':
            bpy.ops.view3d.viewnumpad(type=self.initial_view_orientation)

        self.exit()

    def exit(self):
        preference().BoxCutter_help_item = self.initial_help_item
        bpy.types.SpaceView3D.draw_handler_remove(self.draw_handler, "WINDOW")


    def roll_angle(self, context):
        if self.is_positive:
            angle = -degrees(atan2(self.mouse.y-self.mouse_start.y, self.mouse.x-self.mouse_start.x))

        else:
            angle = degrees(atan2(self.mouse.y-self.mouse_start.y, self.mouse.x-self.mouse_start.x))

        if self.roll_factor:
            angle = round(angle)

            angle = angle - (angle % self.roll_factor) if angle % self.roll_factor < self.roll_factor * 0.5 else (angle - (angle % self.roll_factor) + self.roll_factor)

        return angle


    def roll_view(self, context):
        if self.is_positive:
            self.region_3d.view_rotation = self.initial_rotation * Quaternion(Vector((0, 0, -1)), radians(self.roll_angle))

        else:
            self.region_3d.view_rotation = self.initial_rotation * Quaternion(Vector((0, 0, -1)), radians(-self.roll_angle))

